from collections import namedtuple
import re
from requests.adapters import HTTPAdapter
import socket
from urllib3.connection import HTTPConnection
from urllib3.connectionpool import HTTPConnectionPool

from galaxy_importer.schema import MAX_LENGTH_NAME, MAX_LENGTH_VERSION

CollectionFilename = namedtuple("CollectionFilename", ["namespace", "name", "version"])

LOCALHOST = "localhost"
FILENAME_REGEXP = re.compile(
    r"^(?P<namespace>\w+)-(?P<name>\w+)-" r"(?P<version>[0-9a-zA-Z.+-]+)\.tar\.gz$"
)
VERSION_REGEXP = re.compile(r"""
^
(?P<major>0|[1-9][0-9]*)
\.
(?P<minor>0|[1-9][0-9]*)
\.
(?P<patch>0|[1-9][0-9]*)
(?:
    -(?P<pre>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*)
)?
(?:
    \+(?P<build>[0-9A-Za-z-]+(?:\.[0-9A-Za-z-]+)*)
)?
$
""", re.VERBOSE | re.ASCII)


def parse_collection_filename(filename):
    """
    Parses collection filename.

    Parses and validates collection filename. Returns CollectionFilename named tuple.
    Raises ValueError if filename is not a valid collection filename.
    """
    match = FILENAME_REGEXP.match(filename)

    if not match:
        msg = "Invalid filename {filename}. Expected format: namespace-name-version.tar.gz"
        raise ValueError(msg.format(filename=filename))

    namespace, name, version = match.groups()

    match = VERSION_REGEXP.match(version)
    if not match:
        msg = "Invalid version string {version} from filename {filename}. Expected semantic version format." # noqa
        raise ValueError(msg.format(version=version, filename=filename))

    if len(namespace) > MAX_LENGTH_NAME:
        raise ValueError(f"Expected namespace to be max length of {MAX_LENGTH_NAME}")
    if len(name) > MAX_LENGTH_NAME:
        raise ValueError(f"Expected name to be max length of {MAX_LENGTH_NAME}")
    if len(version) > MAX_LENGTH_VERSION:
        raise ValueError(f"Expected version to be max length of {MAX_LENGTH_VERSION}")

    return CollectionFilename(namespace, name, version)


class SocketHTTPConnection(HTTPConnection):
    def __init__(self, socket_file):
        self.socket_file = socket_file
        super().__init__(LOCALHOST)

    def connect(self):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.socket_file)


class SocketHTTPConnectionPool(HTTPConnectionPool):
    def __init__(self, socket_file):
        self.socket_file = socket_file
        super().__init__(LOCALHOST)

    def _new_conn(self):
        return SocketHTTPConnection(self.socket_file)


class SocketHTTPAdapter(HTTPAdapter):
    def __init__(self, socket_file):
        self.socket_file = socket_file
        super().__init__()

    def get_connection(self, url, proxies=None):
        return SocketHTTPConnectionPool(self.socket_file)



# Generated by Django 2.2.15 on 2020-08-27 17:32

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('galaxy', '0006_collectionsynctask'),
        ('galaxy', '0006_merge_synclist_rbac'),
    ]

    operations = [
    ]



from django.contrib.auth import password_validation
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from galaxy_ng.app.models import auth as auth_models


class GroupSerializer(serializers.ModelSerializer):

    class Meta:
        model = auth_models.Group
        fields = (
            'id',
            'name'
        )


class UserSerializer(serializers.ModelSerializer):

    class Meta:
        model = auth_models.User
        fields = (
            'id',
            'username',
            'first_name',
            'last_name',
            'email',
            'groups',
            'password',
            'date_joined',
            'is_superuser'
        )
        extra_kwargs = {
            'date_joined': {'read_only': True},
            'password': {'write_only': True, 'allow_blank': True, 'required': False}
        }

    def validate_password(self, password):
        if password:
            password_validation.validate_password(password)
            return password

    def validate_groups(self, groups):
        request_user = self.context['request'].user

        group_set = set(groups)
        instance_group_set = set()
        if self.instance:
            instance_group_set = set(list(self.instance.groups.all()))

        group_difference = instance_group_set.symmetric_difference(group_set)

        if not request_user.has_perm('galaxy.change_group'):
            authed_user_groups = request_user.groups.all()
            for g in group_difference:
                if not authed_user_groups.filter(pk=g.id).exists():
                    raise ValidationError(detail={
                        "groups": "'galaxy.change_group' permission is required to change"
                                  " a users group that the requesting user is not in."
                    })

        return groups

    def validate_is_superuser(self, data):
        request_user = self.context['request'].user
        if request_user.is_superuser != data:
            if not request_user.is_superuser:
                raise ValidationError(detail={
                    "is_superuser": "Must be a super user to grant super user permissions."
                })

        return data

    def _set_password(self, instance, data):
        # password doesn't get set the same as other data, so delete it
        # before the serializer saves
        password = data.pop('password', None)
        if password:
            instance.set_password(password)
        return instance

    def create(self, data):
        instance = super().create(data)
        instance = self._set_password(instance, data)
        instance.save()
        return instance

    def update(self, instance, data):
        instance = self._set_password(instance, data)
        return super().update(instance, data)

    def to_representation(self, instance):
        representation = super().to_representation(instance)

        representation['groups'] = GroupSerializer(instance.groups.all(), many=True).data
        return representation

    def to_internal_value(self, data):
        groups = data.get('groups')
        if groups:
            group_ids = []
            for group in groups:
                group_filter = {}
                for field in group:
                    if field in ('id', 'name'):
                        group_filter[field] = group[field]
                try:
                    group = auth_models.Group.objects.get(**group_filter)
                    group_ids.append(group.id)
                except auth_models.Group.DoesNotExist:
                    raise ValidationError(detail={
                        'groups': "Group name=%s, id=%s does not exist" % (group.get('name'),
                                                                           group.get('id'))})
                except ValueError:
                    raise ValidationError(detail={'group': 'Invalid group name or ID'})
            data['groups'] = group_ids
        return super().to_internal_value(data)


class CurrentUserSerializer(UserSerializer):
    model_permissions = serializers.SerializerMethodField()

    class Meta(UserSerializer.Meta):
        model = auth_models.User
        fields = UserSerializer.Meta.fields + ('model_permissions',)
        extra_kwargs = dict(
            groups={'read_only': True},
            **UserSerializer.Meta.extra_kwargs
        )

    def get_model_permissions(self, obj):
        return {
            "add_namespace": obj.has_perm('galaxy.add_namespace'),
            "upload_to_namespace": obj.has_perm('galaxy.upload_to_namespace'),
            "change_namespace": obj.has_perm('galaxy.change_namespace'),
            "move_collection": obj.has_perm('ansible.modify_ansible_repo_content'),
            "change_remote": obj.has_perm('ansible.change_collectionremote'),
            "delete_remote": obj.has_perm('ansible.delete_collectionremote'),
            "add_remote": obj.has_perm('ansible.add_collectionremote'),
            "change_distribution": obj.has_perm('ansible.change_ansibledistribution'),
            "delete_distribution": obj.has_perm('ansible.delete_ansibledistribution'),
            "add_distribution": obj.has_perm('ansible.add_ansibledistribution'),
            "view_distribution": obj.has_perm('ansible.view_ansibledistribution'),
            "view_user": obj.has_perm('galaxy.view_user'),
            "delete_user": obj.has_perm('galaxy.delete_user'),
            "change_user": obj.has_perm('galaxy.change_user'),
            "add_user": obj.has_perm('galaxy.add_user'),
            "add_group": obj.has_perm('galaxy.add_group'),
            "delete_group": obj.has_perm('galaxy.delete_group'),
            "change_group": obj.has_perm('galaxy.change_group'),
            "view_group": obj.has_perm('galaxy.view_group'),
        }



# Generated by Django 2.2.19 on 2021-03-15 18:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('container', '0018_containerdistribution_description'),
        ('galaxy', '0014_containerdistribution'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContainerDistroReadme',
            fields=[
                ('container', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='container.ContainerDistribution')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True, null=True)),
                ('text', models.TextField(blank=True)),
            ],
        ),
    ]



import base64


def user_x_rh_identity(username, account_number=None):
    account_number = account_number or "12345"
    title_username = username.title()
    token_json = """{
        "entitlements":
            {"insights":
                {"is_entitled": true}
            },
        "identity":
            {"account_number": "%(account_number)s",
            "user":
                {"username": "%(username)s",
                "email": "%(username)s",
                "first_name": "%(title_username)s",
                "last_name": "%(title_username)sington",
                "is_org_admin": true
                },
            "internal": {"org_id": "54321"}
            }
            }""" % {'username': username, 'title_username': title_username,
                    'account_number': account_number}

    token_b64 = base64.b64encode(token_json.encode())

    return token_b64



import logging
import re

from django.db import transaction
from django.core import validators

from rest_framework.exceptions import ValidationError
from rest_framework import serializers


from galaxy_ng.app import models
from galaxy_ng.app.access_control.fields import GroupPermissionField

log = logging.getLogger(__name__)


class ScopedErrorListSerializer(serializers.ListSerializer):
    # Updates the list serializer to return error messages as "<childname>__<fieldname>"
    # This is to accomodate for cases where a serializer has to validate a list of
    # sub serializers. Normally error messages will just return the child's field name
    # but this can lead to situations where it's not clear if an error is originating
    # from the child or parent serializer when they share field names.
    def run_validation(self, *args, **kwargs):
        scoped_err_name = self.child.Meta.scoped_error_name

        try:
            return super().run_validation(*args, **kwargs)
        except (ValidationError) as exc:
            new_detail = []
            # loop through list of errors
            for err in exc.detail:
                new_err = {}
                # loop for fields in error
                for field in err:
                    new_err["{}__{}".format(scoped_err_name, field)] = err[field]

                new_detail.append(new_err)

            exc.detail = new_detail
            raise


class NamespaceLinkSerializer(serializers.ModelSerializer):
    # Using a CharField instead of a URLField so that we can add a custom error
    # message that includes the submitted URL
    url = serializers.CharField(
        max_length=256,
        allow_blank=False
    )

    class Meta:
        model = models.NamespaceLink
        fields = ('name', 'url')
        list_serializer_class = ScopedErrorListSerializer
        scoped_error_name = 'links'

    # adds the URL to the error so the user can figure out which link the error
    # message is for
    def validate_url(self, url):
        v = validators.URLValidator(message=f"'{url}' is not a valid url.")
        v(url)
        return url


class NamespaceSerializer(serializers.ModelSerializer):
    links = NamespaceLinkSerializer(many=True, required=False)

    groups = GroupPermissionField()

    class Meta:
        model = models.Namespace
        fields = (
            'id',
            'name',
            'company',
            'email',
            'avatar_url',
            'description',
            'links',
            'groups',
            'resources'
        )

    # replace with a NamespaceNameSerializer and validate_name() ?
    def validate_name(self, name):
        if not name:
            raise ValidationError(detail={
                'name': "Attribute 'name' is required"})
        if not re.match(r'^[a-z0-9_]+$', name):
            raise ValidationError(detail={
                'name': 'Name can only contain lower case letters, underscores and numbers'})
        if len(name) <= 2:
            raise ValidationError(detail={
                'name': 'Name must be longer than 2 characters'})
        if name.startswith('_'):
            raise ValidationError(detail={
                'name': "Name cannot begin with '_'"})
        return name

    @transaction.atomic
    def create(self, validated_data):
        links_data = validated_data.pop('links', [])

        instance = models.Namespace.objects.create(**validated_data)

        # create NamespaceLink objects if needed
        new_links = []
        for link_data in links_data:
            link_data["namespace"] = instance
            ns_link, created = models.NamespaceLink.objects.get_or_create(**link_data)
            new_links.append(ns_link)

        instance.links.set(new_links)
        return instance

    @transaction.atomic
    def update(self, instance, validated_data):
        links = validated_data.pop('links', None)

        if links is not None:
            instance.set_links(links)

        instance = super().update(instance, validated_data)
        instance.save()
        return instance


class NamespaceSummarySerializer(NamespaceSerializer):
    """NamespaceSerializer but without 'links' or 'resources'.

    For use in _ui/collection detail views."""

    class Meta:
        model = models.Namespace
        fields = (
            'id',
            'name',
            'company',
            'email',
            'avatar_url',
            'description',
            'groups',
        )

        read_only_fields = ('name', )



import logging
from typing import List

from django.conf import settings
from rest_access_policy import AccessPolicy, AccessPolicyException
from rest_framework.exceptions import NotFound

from galaxy_ng.app import models

from galaxy_ng.app.access_control.statements import STANDALONE_STATEMENTS, INSIGHTS_STATEMENTS

log = logging.getLogger(__name__)

STATEMENTS = {'insights': INSIGHTS_STATEMENTS,
              'standalone': STANDALONE_STATEMENTS}


class AccessPolicyVerboseMixin:
    """Mixin to make AccessPolicy checks verbose

    This has to reimplement / cut&paste most of rest_access_policy.AccessPolicy
    to instrument the low level permissions checks.
    """

    def has_permission(self, request, view):
        result = super().has_permission(request, view)
        log.debug('"%s" perm check was %s for "%s %s" view="%s-%s-%s" for user="%s" with groups=%s',
                  self.NAME,
                  result,
                  request._request.method,
                  request._request.path,
                  getattr(view, 'basename', 'NotAViewSet'),
                  getattr(view, 'action', 'NotAViewSet'),
                  getattr(view, 'detail', 'NotAViewSet'),
                  request.user, ','.join([x.name for x in request.user.groups.all()]),
                  )
        return result

    def has_object_permission(self, request, view, obj):
        result = super().has_object_permission(request, view, obj)
        log.debug('"%s" %s %s view=%s-%s-%s for user=%s, groups=%s obj=%s had result: %s',
                  self.NAME,
                  request._request.method,
                  request._request.path,
                  view.basename, view.action, view.detail,
                  request.user,
                  ','.join([x.name for x in request.user.groups.all()]),
                  obj,
                  result)
        return result

    def _evaluate_statements(
        self, statements: List[dict], request, view, action: str
    ) -> bool:

        statements = self._normalize_statements(statements)

        user = request.user
        user_groups = self.get_user_group_values(user)

        matched = self._get_statements_matching_principal(request, statements)
        matched_principals = set()
        for match in matched:
            for principal in match['principal']:
                matched_principals.add(principal)

        log.debug('"%s" user "%s" in groups %s matched access policy principals %s',
                  self.NAME,
                  request.user,
                  ",".join(['"%s"' % ug for ug in user_groups]),
                  matched_principals)

        matched = self._get_statements_matching_action(request, action, matched)

        log.debug('"%s" action "%s" matched statements %s',
                  self.NAME, action, matched)

        matched = self._get_statements_matching_context_conditions(
            request, view, action, matched
        )

        denied = [_ for _ in matched if _["effect"] != "allow"]

        if len(matched) == 0 or len(denied) > 0:
            return False

        return True

    def _check_condition(self, condition: str, request, view, action: str):
        """
            Evaluate a custom context condition; if method does not exist on
            the access policy class, then return False.
            Condition value can contain a value that is passed to method, if
            formatted as `<method_name>:<arg_value>`.
        """

        parts = condition.split(":", 1)
        method_name = parts[0]
        arg = parts[1] if len(parts) == 2 else None

        method = self._get_condition_method(method_name)

        if arg is not None:
            result = method(request, view, action, arg)
        else:
            result = method(request, view, action)

        if type(result) is not bool:
            raise AccessPolicyException(
                "condition '%s' must return true/false, not %s"
                % (condition, type(result))
            )

        res_blurb = "failed"
        if result:
            res_blurb = "passed"

        log.debug('"%s" action "%s" for user "%s" %s conditions "%s"',
                  self.NAME,
                  action,
                  request.user,
                  res_blurb,
                  condition,
                  )

        return result

    def _get_statements_matching_principal(
        self, request, statements: List[dict]
    ) -> List[dict]:
        user = request.user
        user_roles = None
        matched = []

        for statement in statements:
            principals = statement["principal"]
            found = False

            if "*" in principals:
                found = True
            elif "authenticated" in principals:
                found = not user.is_anonymous
            elif "anonymous" in principals:
                found = user.is_anonymous
            elif self.id_prefix + str(user.pk) in principals:
                found = True
            else:
                log.debug("No '*', 'authenticated', 'anonymous', or user id in %s,"
                          + "trying groups %s",
                          principals,
                          user_roles)

                if not user_roles:
                    user_roles = self.get_user_group_values(user)

                for user_role in user_roles:
                    if self.group_prefix + user_role in principals:
                        found = True
                        break

            if found:
                matched.append(statement)

        return matched


class AccessPolicyBase(AccessPolicyVerboseMixin, AccessPolicy):
    def _get_statements(self, deployment_mode):
        return STATEMENTS[deployment_mode]

    def get_policy_statements(self, request, view):
        statements = self._get_statements(settings.GALAXY_DEPLOYMENT_MODE)
        return statements.get(self.NAME, [])

    def _get_rh_identity(self, request):
        if not isinstance(request.auth, dict):
            log.debug("No request rh_identity request.auth found for request %s", request)
            return False

        x_rh_identity = request.auth.get('rh_identity')
        if not x_rh_identity:
            return False

        return x_rh_identity

    # used by insights access policy
    def has_rh_entitlements(self, request, view, permission):

        x_rh_identity = self._get_rh_identity(request)

        if not x_rh_identity:
            log.debug("No x_rh_identity found when check entitlements for request %s for view %s",
                      request, view)
            return False

        entitlements = x_rh_identity.get('entitlements', {})
        entitlement = entitlements.get(settings.RH_ENTITLEMENT_REQUIRED, {})
        return entitlement.get('is_entitled', False)

    def is_org_admin(self, request, view, permission):
        """Check the rhn_entitlement data to see if user is an org admin"""
        x_rh_identity = self._get_rh_identity(request)
        if not x_rh_identity:
            log.debug("No x_rh_identity found for request %s for view %s",
                      request, view)
            return False

        identity = x_rh_identity['identity']
        user = identity['user']
        return user.get('is_org_admin', False)


class NamespaceAccessPolicy(AccessPolicyBase):
    NAME = 'NamespaceViewSet'


class CollectionAccessPolicy(AccessPolicyBase):
    NAME = 'CollectionViewSet'

    def can_update_collection(self, request, view, permission):
        collection = view.get_object()
        namespace = models.Namespace.objects.get(name=collection.namespace)
        return request.user.has_perm('galaxy.upload_to_namespace', namespace)

    def can_create_collection(self, request, view, permission):
        data = view._get_data(request)
        try:
            namespace = models.Namespace.objects.get(name=data['filename'].namespace)
        except models.Namespace.DoesNotExist:
            raise NotFound('Namespace in filename not found.')
        return request.user.has_perm('galaxy.upload_to_namespace', namespace)


class CollectionRemoteAccessPolicy(AccessPolicyBase):
    NAME = 'CollectionRemoteViewSet'


class UserAccessPolicy(AccessPolicyBase):
    NAME = 'UserViewSet'

    def user_is_superuser(self, request, view, action):
        user = view.get_object()
        return user.is_superuser

    def is_current_user(self, request, view, action):
        return request.user == view.get_object()


class MyUserAccessPolicy(AccessPolicyBase):
    NAME = 'MyUserViewSet'

    def is_current_user(self, request, view, action):
        return request.user == view.get_object()


class SyncListAccessPolicy(AccessPolicyBase):
    NAME = 'SyncListViewSet'


class MySyncListAccessPolicy(AccessPolicyBase):
    NAME = 'MySyncListViewSet'


class TagsAccessPolicy(AccessPolicyBase):
    NAME = 'TagViewSet'


class TaskAccessPolicy(AccessPolicyBase):
    NAME = 'TaskViewSet'


class LoginAccessPolicy(AccessPolicyBase):
    NAME = 'LoginView'


class LogoutAccessPolicy(AccessPolicyBase):
    NAME = 'LogoutView'


class TokenAccessPolicy(AccessPolicyBase):
    NAME = 'TokenView'


class GroupAccessPolicy(AccessPolicyBase):
    NAME = 'GroupViewSet'


class DistributionAccessPolicy(AccessPolicyBase):
    NAME = 'DistributionViewSet'


class MyDistributionAccessPolicy(AccessPolicyBase):
    NAME = 'MyDistributionViewSet'


class ContainerRepositoryAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerRepositoryViewSet'


class ContainerReadmeAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerReadmeViewset'

    def has_container_namespace_perms(self, request, view, action, permission):
        readme = view.get_object()
        return (request.user.has_perm(permission)
                or request.user.has_perm(permission, readme.container.namespace))


class ContainerNamespaceAccessPolicy(AccessPolicyBase):
    NAME = 'ContainerNamespaceViewset'



"""Utilities for tests for the galaxy plugin."""
from functools import partial
import requests
from unittest import SkipTest
from time import sleep
from tempfile import NamedTemporaryFile

from pulp_smash import api, selectors
from pulp_smash.pulp3.utils import (
    gen_remote,
    gen_repo,
    get_content,
    require_pulp_3,
    require_pulp_plugins,
    sync,
)

from galaxy_ng.tests.functional.constants import (
    GALAXY_CONTENT_NAME,
    GALAXY_CONTENT_PATH,
    GALAXY_FIXTURE_URL,
    GALAXY_PUBLICATION_PATH,
    GALAXY_REMOTE_PATH,
    GALAXY_REPO_PATH,
    GALAXY_URL,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    ArtifactsApi,
    Configuration,
    TasksApi,
)
from pulpcore.client.galaxy_ng import ApiClient as GalaxyApiClient


configuration = Configuration()
configuration.username = "admin"
configuration.password = "password"
configuration.safe_chars_for_path_param = "/"


def set_up_module():
    """Skip tests Pulp 3 isn't under test or if galaxy_ng isn't installed."""
    require_pulp_3(SkipTest)
    require_pulp_plugins({"galaxy_ng"}, SkipTest)


def gen_galaxy_client():
    """Return an OBJECT for galaxy client."""
    return GalaxyApiClient(configuration)


def gen_galaxy_remote(url=GALAXY_FIXTURE_URL, **kwargs):
    """Return a semi-random dict for use in creating a galaxy Remote.

    :param url: The URL of an external content source.
    """
    # FIXME: Add any fields specific to a galaxy remote here
    return gen_remote(url, **kwargs)


def get_galaxy_content_paths(repo, version_href=None):
    """Return the relative path of content units present in a galaxy repository.

    :param repo: A dict of information about the repository.
    :param version_href: The repository version to read.
    :returns: A dict of lists with the paths of units present in a given repository.
        Paths are given as pairs with the remote and the local version for different content types.
    """
    # FIXME: The "relative_path" is actually a file path and name
    # It's just an example -- this needs to be replaced with an implementation that works
    # for repositories of this content type.
    return {
        GALAXY_CONTENT_NAME: [
            (content_unit["relative_path"], content_unit["relative_path"])
            for content_unit in get_content(repo, version_href)[GALAXY_CONTENT_NAME]
        ],
    }


def gen_galaxy_content_attrs(artifact):
    """Generate a dict with content unit attributes.

    :param artifact: A dict of info about the artifact.
    :returns: A semi-random dict for use in creating a content unit.
    """
    # FIXME: Add content specific metadata here.
    return {"_artifact": artifact["pulp_href"]}


def populate_pulp(cfg, url=GALAXY_FIXTURE_URL):
    """Add galaxy contents to Pulp.

    :param pulp_smash.config.PulpSmashConfig: Information about a Pulp application.
    :param url: The galaxy repository URL. Defaults to
        :data:`pulp_smash.constants.GALAXY_FIXTURE_URL`
    :returns: A list of dicts, where each dict describes one galaxy content in Pulp.
    """
    client = api.Client(cfg, api.json_handler)
    remote = {}
    repo = {}
    try:
        remote.update(client.post(GALAXY_REMOTE_PATH, gen_galaxy_remote(url)))
        repo.update(client.post(GALAXY_REPO_PATH, gen_repo()))
        sync(cfg, remote, repo)
    finally:
        if remote:
            client.delete(remote["pulp_href"])
        if repo:
            client.delete(repo["pulp_href"])
    return client.get(GALAXY_CONTENT_PATH)["results"]


def publish(cfg, repo, version_href=None):
    """Publish a repository.
    :param pulp_smash.config.PulpSmashConfig cfg: Information about the Pulp
        host.
    :param repo: A dict of information about the repository.
    :param version_href: A href for the repo version to be published.
    :returns: A publication. A dict of information about the just created
        publication.
    """
    if version_href:
        body = {"repository_version": version_href}
    else:
        body = {"repository": repo["pulp_href"]}

    client = api.Client(cfg, api.json_handler)
    call_report = client.post(GALAXY_PUBLICATION_PATH, body)
    tasks = tuple(api.poll_spawned_tasks(cfg, call_report))
    return client.get(tasks[-1]["created_resources"][0])


skip_if = partial(selectors.skip_if, exc=SkipTest)  # pylint:disable=invalid-name
"""The ``@skip_if`` decorator, customized for unittest.

:func:`pulp_smash.selectors.skip_if` is test runner agnostic. This function is
identical, except that ``exc`` has been set to ``unittest.SkipTest``.
"""

core_client = CoreApiClient(configuration)
tasks = TasksApi(core_client)


def gen_artifact(url=GALAXY_URL):
    """Creates an artifact."""
    response = requests.get(url)
    with NamedTemporaryFile() as temp_file:
        temp_file.write(response.content)
        temp_file.flush()
        artifact = ArtifactsApi(core_client).create(file=temp_file.name)
        return artifact.to_dict()


def monitor_task(task_href):
    """Polls the Task API until the task is in a completed state.

    Prints the task details and a success or failure message. Exits on failure.

    Args:
        task_href(str): The href of the task to monitor

    Returns:
        list[str]: List of hrefs that identify resource created by the task

    """
    completed = ["completed", "failed", "canceled"]
    task = tasks.read(task_href)
    while task.state not in completed:
        sleep(2)
        task = tasks.read(task_href)

    if task.state == "completed":
        return task.created_resources

    return task.to_dict()



from galaxy_ng.app.constants import COMMUNITY_DOMAINS
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.exceptions import ValidationError

from django.shortcuts import get_object_or_404

from pulp_ansible.app import models as pulp_models
from pulp_ansible.app.tasks.collections import sync as collection_sync

from pulpcore.plugin.tasking import enqueue_with_reservation
from pulpcore.plugin.models import Task

from galaxy_ng.app.api import base as api_base
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app import models


class SyncRemoteView(api_base.APIView):
    permission_classes = [access_policy.CollectionRemoteAccessPolicy]
    action = 'sync'

    def post(self, request: Request, *args, **kwargs) -> Response:
        distro_path = kwargs['path']
        distro = get_object_or_404(pulp_models.AnsibleDistribution, base_path=distro_path)

        if not distro.repository or not distro.repository.remote:
            raise ValidationError(
                detail={'remote': f'The {distro_path} distribution does not have'
                                  ' any remotes associated with it.'})

        remote = distro.repository.remote.ansible_collectionremote

        if not remote.requirements_file and any(
            [domain in remote.url for domain in COMMUNITY_DOMAINS]
        ):
            raise ValidationError(
                detail={
                    'requirements_file':
                        'Syncing content from galaxy.ansible.com without specifying a '
                        'requirements file is not allowed.'
                })

        result = enqueue_with_reservation(
            collection_sync,
            [distro.repository, remote],
            kwargs={
                "remote_pk": remote.pk,
                "repository_pk": distro.repository.pk,
                "mirror": True,
                "optimize": True,
            },
        )

        repo = pulp_models.AnsibleRepository.objects.get(pk=distro.repository.pk)
        task = Task.objects.get(pk=result.id)

        models.CollectionSyncTask.objects.create(
            repository=repo,
            task=task
        )

        return Response({'task': task.pk})



import logging

import requests

from django.core.exceptions import ValidationError, ObjectDoesNotExist


from django.conf import settings
from django.http import HttpResponseRedirect, StreamingHttpResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse

from rest_framework.response import Response
from rest_framework.exceptions import APIException, NotFound

from pulpcore.plugin.models import Task
from pulpcore.plugin.tasking import enqueue_with_reservation
from pulp_ansible.app.galaxy.v3 import views as pulp_ansible_views
from pulp_ansible.app.models import CollectionVersion, AnsibleDistribution
from pulp_ansible.app.models import CollectionImport as PulpCollectionImport
from galaxy_ng.app.api import base as api_base

from galaxy_ng.app.constants import DeploymentMode, INBOUND_REPO_NAME_FORMAT
from galaxy_ng.app import models
from galaxy_ng.app.access_control import access_policy

from galaxy_ng.app.api.utils import SocketHTTPAdapter
from galaxy_ng.app.api.v3.serializers import (
    CollectionSerializer,
    CollectionVersionSerializer,
    UnpaginatedCollectionVersionSerializer,
    CollectionVersionListSerializer,
    CollectionUploadSerializer,
)

from galaxy_ng.app.common import metrics
from galaxy_ng.app.tasks import (
    import_and_move_to_staging,
    import_and_auto_approve,
    call_copy_task,
    call_remove_task,
    curate_all_synclist_repository,
)

from galaxy_ng.app.common.parsers import AnsibleGalaxy29MultiPartParser


log = logging.getLogger(__name__)


class ViewNamespaceSerializerContextMixin:
    def get_serializer_context(self):
        """Inserts distribution path to a serializer context."""

        context = super().get_serializer_context()

        # view_namespace will be used by the serializers that need to return different hrefs
        # depending on where in the urlconf they are.
        # view_route is the url 'route' pattern, used to
        # handle the special case /api/automation-hub/v3/collections/ not having
        # a <str:path> in it's url
        request = context.get("request", None)
        context["view_namespace"] = None
        if request:
            context["view_namespace"] = request.resolver_match.namespace
            context["view_route"] = request.resolver_match.route

        return context


class RepoMetadataViewSet(api_base.LocalSettingsMixin,
                          pulp_ansible_views.RepoMetadataViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]


class UnpaginatedCollectionViewSet(api_base.LocalSettingsMixin,
                                   ViewNamespaceSerializerContextMixin,
                                   pulp_ansible_views.UnpaginatedCollectionViewSet):
    pagination_class = None
    permission_classes = [access_policy.CollectionAccessPolicy]
    serializer_class = CollectionSerializer


class CollectionViewSet(api_base.LocalSettingsMixin,
                        ViewNamespaceSerializerContextMixin,
                        pulp_ansible_views.CollectionViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]
    serializer_class = CollectionSerializer


class UnpaginatedCollectionVersionViewSet(api_base.LocalSettingsMixin,
                                          ViewNamespaceSerializerContextMixin,
                                          pulp_ansible_views.UnpaginatedCollectionVersionViewSet):
    pagination_class = None
    serializer_class = UnpaginatedCollectionVersionSerializer
    permission_classes = [access_policy.CollectionAccessPolicy]


class CollectionVersionViewSet(api_base.LocalSettingsMixin,
                               ViewNamespaceSerializerContextMixin,
                               pulp_ansible_views.CollectionVersionViewSet):
    serializer_class = CollectionVersionSerializer
    permission_classes = [access_policy.CollectionAccessPolicy]
    list_serializer_class = CollectionVersionListSerializer


class CollectionVersionDocsViewSet(api_base.LocalSettingsMixin,
                                   pulp_ansible_views.CollectionVersionDocsViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]


class CollectionImportViewSet(api_base.LocalSettingsMixin,
                              pulp_ansible_views.CollectionImportViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]


class CollectionUploadViewSet(api_base.LocalSettingsMixin,
                              pulp_ansible_views.CollectionUploadViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]
    parser_classes = [AnsibleGalaxy29MultiPartParser]
    serializer_class = CollectionUploadSerializer

    def _dispatch_import_collection_task(self, temp_file_pk, repository=None, **kwargs):
        """Dispatch a pulp task started on upload of collection version."""
        locks = []

        kwargs["temp_file_pk"] = temp_file_pk

        if repository:
            locks.append(repository)
            kwargs["repository_pk"] = repository.pk

        if settings.GALAXY_REQUIRE_CONTENT_APPROVAL:
            return enqueue_with_reservation(import_and_move_to_staging, locks, kwargs=kwargs)
        return enqueue_with_reservation(import_and_auto_approve, locks, kwargs=kwargs)

    # Wrap super().create() so we can create a galaxy_ng.app.models.CollectionImport based on the
    # the import task and the collection artifact details

    def _get_data(self, request):
        serializer = CollectionUploadSerializer(data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)

        return serializer.validated_data

    @staticmethod
    def _get_path(kwargs, filename_ns):
        """Use path from '/content/<path>/v3/' or
           if user does not specify distribution base path
           then use an inbound distribution based on filename namespace.
        """
        path = kwargs['path']
        if kwargs.get('no_path_specified', None):
            path = INBOUND_REPO_NAME_FORMAT.format(namespace_name=filename_ns)
        return path

    @staticmethod
    def _check_path_matches_expected_repo(path, filename_ns):
        """Reject if path does not match expected inbound format
           containing filename namespace.

        Examples:
        Reject if path is "staging".
        Reject if path does not start with "inbound-".
        Reject if path is "inbound-alice" but filename namepace is "bob".
        """

        distro = get_object_or_404(AnsibleDistribution, base_path=path)
        repo_name = distro.repository.name
        if INBOUND_REPO_NAME_FORMAT.format(namespace_name=filename_ns) == repo_name:
            return
        raise NotFound(
            f'Path does not match: "{INBOUND_REPO_NAME_FORMAT.format(namespace_name=filename_ns)}"')

    def create(self, request, *args, **kwargs):
        data = self._get_data(request)
        filename = data['filename']

        path = self._get_path(kwargs, filename_ns=filename.namespace)

        try:
            namespace = models.Namespace.objects.get(name=filename.namespace)
        except models.Namespace.DoesNotExist:
            raise ValidationError(
                'Namespace "{0}" does not exist.'.format(filename.namespace)
            )

        self._check_path_matches_expected_repo(path, filename_ns=namespace.name)

        self.check_object_permissions(request, namespace)

        try:
            response = super(CollectionUploadViewSet, self).create(request, path)
        except ValidationError:
            log.exception('Failed to publish artifact %s (namespace=%s, sha256=%s)',  # noqa
                          data['file'].name, namespace, data.get('sha256'))
            raise

        task_href = response.data['task']

        # icky, have to extract task id from the task_href url
        task_id = task_href.strip("/").split("/")[-1]

        task_detail = Task.objects.get(pk=task_id)

        pulp_collection_import = PulpCollectionImport.objects.get(pk=task_id)

        models.CollectionImport.objects.create(
            task_id=pulp_collection_import,
            created_at=task_detail.pulp_created,
            namespace=namespace,
            name=data['filename'].name,
            version=data['filename'].version,
        )

        # TODO: CollectionImport.get_absolute_url() should be able to generate this, but
        #       it needs the  repo/distro base_path for the <path> part of url
        import_obj_url = reverse("galaxy:api:content:v3:collection-import",
                                 kwargs={'pk': str(task_detail.pulp_id),
                                         'path': path})

        log.debug('import_obj_url: %s', import_obj_url)
        return Response(
            data={'task': import_obj_url},
            status=response.status_code
        )


class CollectionArtifactDownloadView(api_base.APIView):
    permission_classes = [access_policy.CollectionAccessPolicy]
    action = 'retrieve'

    def _get_tcp_response(self, url):
        return requests.get(url, stream=True, allow_redirects=False)

    def _get_unix_socket_response(self, url):
        socket_file = settings.CONTENT_BIND.split(':')[1]
        session = requests.Session()
        session.mount("http://", SocketHTTPAdapter(socket_file))
        return session.get(url, stream=True, allow_redirects=False)

    def get(self, request, *args, **kwargs):
        metrics.collection_artifact_download_attempts.inc()

        url = 'http://{host}:{port}/{prefix}/{distro_base_path}/{filename}'.format(
            host=settings.X_PULP_CONTENT_HOST,
            port=settings.X_PULP_CONTENT_PORT,
            prefix=settings.CONTENT_PATH_PREFIX.strip('/'),
            distro_base_path=self.kwargs['path'],
            filename=self.kwargs['filename'],
        )

        content_bind = settings.get("CONTENT_BIND", None)
        if content_bind and content_bind.startswith("unix:"):
            response = self._get_unix_socket_response(url)
        else:
            response = self._get_tcp_response(url)

        if response.status_code == requests.codes.not_found:
            metrics.collection_artifact_download_failures.labels(
                status=requests.codes.not_found
            ).inc()
            raise NotFound()

        if response.status_code == requests.codes.found:
            return HttpResponseRedirect(response.headers['Location'])

        if response.status_code == requests.codes.ok:
            metrics.collection_artifact_download_successes.inc()

            return StreamingHttpResponse(
                response.raw.stream(amt=4096),
                content_type=response.headers['Content-Type']
            )

        metrics.collection_artifact_download_failures.labels(status=response.status_code).inc()
        raise APIException('Unexpected response from content app. '
                           f'Code: {response.status_code}.')


class CollectionVersionMoveViewSet(api_base.ViewSet):
    permission_classes = [access_policy.CollectionAccessPolicy]

    def move_content(self, request, *args, **kwargs):
        """Remove content from source repo and add to destination repo.

        Creates new RepositoryVersion of source repo without content included.
        Creates new RepositoryVersion of destination repo with content included.
        """

        version_str = '-'.join([self.kwargs[key] for key in ('namespace', 'name', 'version')])
        try:
            collection_version = CollectionVersion.objects.get(
                namespace=self.kwargs['namespace'],
                name=self.kwargs['name'],
                version=self.kwargs['version'],
            )
        except ObjectDoesNotExist:
            raise NotFound(f'Collection {version_str} not found')

        try:
            src_repo = AnsibleDistribution.objects.get(
                base_path=self.kwargs['source_path']).repository
            dest_repo = AnsibleDistribution.objects.get(
                base_path=self.kwargs['dest_path']).repository
        except ObjectDoesNotExist:
            raise NotFound(f'Repo(s) for moving collection {version_str} not found')

        src_versions = CollectionVersion.objects.filter(pk__in=src_repo.latest_version().content)
        if collection_version not in src_versions:
            raise NotFound(f'Collection {version_str} not found in source repo')

        dest_versions = CollectionVersion.objects.filter(pk__in=dest_repo.latest_version().content)
        if collection_version in dest_versions:
            raise NotFound(f'Collection {version_str} already found in destination repo')

        copy_task = call_copy_task(collection_version, src_repo, dest_repo)
        remove_task = call_remove_task(collection_version, src_repo)

        curate_task_id = None
        if settings.GALAXY_DEPLOYMENT_MODE == DeploymentMode.INSIGHTS.value:
            golden_repo = AnsibleDistribution.objects.get(
                base_path=settings.GALAXY_API_DEFAULT_DISTRIBUTION_BASE_PATH
            ).repository

            if dest_repo == golden_repo or src_repo == golden_repo:
                repo_name = golden_repo.name
                locks = [golden_repo]
                task_args = (repo_name,)
                task_kwargs = {}

                curate_task = enqueue_with_reservation(
                    curate_all_synclist_repository, locks, args=task_args, kwargs=task_kwargs
                )
                curate_task_id = curate_task.id

        return Response(
            data={
                'copy_task_id': copy_task.id,
                'remove_task_id': remove_task.id,
                "curate_all_synclist_repository_task_id": curate_task_id,
            },
            status='202'
        )



from django.apps import apps
from django.http import HttpResponseRedirect

from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse

from galaxy_ng.app.api import base as api_base


class ApiRootView(api_base.APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request, *args, **kwargs):
        data = {
            "available_versions": {"v3": "v3/"},
            "server_version": apps.get_app_config("galaxy").version,
            "galaxy_ng_version": apps.get_app_config("galaxy").version,
            "pulp_ansible_version": apps.get_app_config('ansible').version,
        }

        if kwargs.get("path"):
            data["distro_base_path"] = kwargs["path"]

        return Response(data)


class ApiRedirectView(api_base.APIView):
    permission_classes = [IsAuthenticated]

    """Redirect requests to /api/automation-hub/api/ to /api/automation-hub/

    This is a workaround for https://github.com/ansible/ansible/issues/62073.
    This can be removed when ansible-galaxy stops appending '/api' to the url."""

    def get(self, request, *args, **kwargs):
        reverse_url_name = kwargs.get("reverse_url_name")

        reverse_kwargs = {}
        if "path" in kwargs:
            reverse_kwargs["path"] = kwargs["path"]

        return HttpResponseRedirect(reverse(reverse_url_name,
                                            kwargs=reverse_kwargs), status=307)



#!/usr/bin/env python3

import os
import re
import tempfile
import tarfile
import shutil
import urllib.request
import urllib.error
from distutils import log

from setuptools import find_packages, setup, Command
from setuptools.command.build_py import build_py as _BuildPyCommand
from setuptools.command.sdist import sdist as _SDistCommand

package_name = os.environ.get("GALAXY_NG_ALTERNATE_NAME", "galaxy-ng")
version = "4.3.3"


class PrepareStaticCommand(Command):
    DEV_UI_DOWNLOAD_URL = (
        "https://github.com/ansible/ansible-hub-ui/"
        "releases/download/dev/automation-hub-ui-dist.tar.gz")

    ALTERNATE_UI_DOWNLOAD_URL = os.environ.get("ALTERNATE_UI_DOWNLOAD_URL")

    UI_DOWNLOAD_URL = (
        "https://github.com/ansible/ansible-hub-ui/"
        f"releases/download/{version}/automation-hub-ui-dist.tar.gz"
    )
    TARGET_DIR = "galaxy_ng/app/static/galaxy_ng"

    user_options = [
        (
            'force-download-ui',
            None,
            'Replace any existing static files with the ones downloaded from github.'
        ),
    ]

    def initialize_options(self):
        self.force_download_ui = False

    def finalize_options(self):
        pass

    def run(self):
        if os.path.exists(self.TARGET_DIR):
            if self.force_download_ui:
                log.warn(f"Removing {self.TARGET_DIR} and re downloading the UI.")
                shutil.rmtree(self.TARGET_DIR)
            else:
                log.warn(f"Static directory {self.TARGET_DIR} already exists, skipping. ")
                return

        with tempfile.NamedTemporaryFile() as download_file:
            log.info(f"Downloading UI distribution to temporary file: {download_file.name}")

            if self.ALTERNATE_UI_DOWNLOAD_URL:
                log.info(f"Downloading UI from {self.ALTERNATE_UI_DOWNLOAD_URL}")
                self._download_tarball(self.ALTERNATE_UI_DOWNLOAD_URL, download_file)
            else:
                log.info(f"Attempting to download UI for version {version}")
                try:
                    self._download_tarball(self.UI_DOWNLOAD_URL, download_file)
                except urllib.error.HTTPError:
                    log.warn(f"Failed to retrieve UI for {version}. Downloading latest UI.")
                    self._download_tarball(self.DEV_UI_DOWNLOAD_URL, download_file)

            log.info(f"Extracting UI static files to {self.TARGET_DIR}")
            with tarfile.open(fileobj=download_file) as tfp:
                tfp.extractall(self.TARGET_DIR)

    def _download_tarball(self, url, download_file):
        urllib.request.urlretrieve(url, filename=download_file.name)


class SDistCommand(_SDistCommand):
    def run(self):
        self.run_command('prepare_static')
        return super().run()


class BuildPyCommand(_BuildPyCommand):
    def run(self):
        self.run_command('prepare_static')
        return super().run()


requirements = [
    "Django~=2.2.23",
    "galaxy-importer==0.3.4",
    "pulpcore<3.12,>=3.11.2",
    "pulp-ansible==0.7.3",
    "django-prometheus>=2.0.0",
    "drf-spectacular",
    # pulp-container 2.6 requires pulpcore >=3.12.1
    "pulp-container>=2.5.2,<2.6.0",
    # click 8 requires py38,
    # can be removed once we require >=py38
    "click==7.1.2",
]


is_on_dev_environment = (
    "COMPOSE_PROFILE" in os.environ and "DEV_SOURCE_PATH" in os.environ
    and os.environ.get("LOCK_REQUIREMENTS") == "0"
)
if is_on_dev_environment:
    """
    To enable the installation of local dependencies e.g: a local fork of
    pulp_ansible checked out to specific branch/version.
    The paths listed on DEV_SOURCE_PATH must be unpinned to avoid pip
    VersionConflict error.
    ref: https://github.com/ansible/galaxy_ng/wiki/Development-Setup
         #steps-to-run-dev-environment-with-specific-upstream-branch
    """
    DEV_SOURCE_PATH = os.environ.get("DEV_SOURCE_PATH", "").split(":")
    DEV_SOURCE_PATH += [path.replace("_", "-") for path in DEV_SOURCE_PATH]
    requirements = [
        re.sub(r"""(=|^|~|<|>|!)([\S]+)""", "", req)
        if req.lower().startswith(tuple(DEV_SOURCE_PATH)) else req
        for req in requirements
    ]
    print("Installing with unpinned DEV_SOURCE_PATH requirements", requirements)

setup(
    name=package_name,
    version=version,
    description="galaxy-ng plugin for the Pulp Project",
    license="GPLv2+",
    author="Red Hat, Inc.",
    author_email="info@ansible.com",
    url="https://github.com/ansible/galaxy_ng/",
    python_requires=">=3.6",
    setup_requires=["wheel"],
    install_requires=requirements,
    include_package_data=True,
    packages=find_packages(exclude=["tests", "tests.*"]),
    classifiers=(
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ),
    entry_points={"pulpcore.plugin": ["galaxy_ng = galaxy_ng:default_app_config"]},
    cmdclass={
        "prepare_static": PrepareStaticCommand,
        "build_py": BuildPyCommand,
        "sdist": SDistCommand,
    },
)



import logging

log = logging.getLogger(__name__)
log.debug('openapi autoschema')


def preprocess_debug_logger(endpoints, **kwargs):
    """Log and then return an iterable of (path, path_regex, method, callback) tuples"""

    log.debug("kwargs: %s", repr(kwargs))

    for path, path_regex, method, callback in endpoints:
        log.debug('path=%s, path_regex=%s, method=%s, callback=%s',
                  path, path_regex, method, callback)

    return endpoints
    # return [
    #     (path, path_regex, method, callback) for path, path_regex, method, callback in endpoints
    #     if log.debug('path=%s, path_regex=%s, method=%s, callback=%s',
    #                  path, path_regex, method, callback)
    # ]


def preprocess_exclude_endpoints(endpoints, **kwargs):
    """Return an iterable of (path, path_regex, method, callback) with some endpoints removed

    For example, the default is to to remove '/pulp' and '/_ui/' api endpoints, while
    allowing /pulp/api/v3/status endpoint and all automation-hub v3 endpoints.
    """
    return [
        (path, path_regex, method, callback) for path, path_regex, method, callback in endpoints
        if path == '/pulp/api/v3/status/' or (not path.startswith('/pulp') and '/_ui/' not in path)
    ]



import tempfile
from io import StringIO

from django.core.management import CommandError, call_command
from django.test import TestCase
from pulp_ansible.app.models import (
    AnsibleDistribution,
    AnsibleRepository,
    CollectionRemote,
)

REQUIREMENTS_FILE_STR = (
    "collections:\n"
    "  - name: initial.name\n"
    "    server: https://initial.content.com/v3/collections/\n"
    "    api_key: NotASecret\n"
)


class TestCreateRemoteCommand(TestCase):

    def setUp(self):
        super().setUp()
        CollectionRemote.objects.create(
            name="test-remote",
            url="https://test.remote/v3/collections/"
        )

    def test_command_output(self):
        with self.assertRaisesMessage(
            CommandError, "Error: the following arguments are required: name, url"
        ):
            call_command('create-remote')

    def test_remote_already_exists(self):
        out = StringIO()

        call_command(
            'create-remote',
            'test-remote',
            'https://test.remote/v3/collections/',
            stdout=out
        )

        self.assertIn('Updated existing CollectionRemote test-remote', out.getvalue())
        self.assertIn('Created new Repository test-remote', out.getvalue())
        self.assertIn('Created new Distribution test-remote', out.getvalue())

    def test_remote_created(self):
        out = StringIO()
        call_command('create-remote', 'new-remote', 'https://new.remote/', stdout=out)

        self.assertIn('Created new CollectionRemote new-remote', out.getvalue())
        self.assertIn('Created new Repository new-remote', out.getvalue())
        self.assertIn('Created new Distribution new-remote', out.getvalue())

        self.assertTrue(CollectionRemote.objects.filter(name='new-remote').exists())
        self.assertTrue(AnsibleRepository.objects.filter(name='new-remote').exists())
        self.assertTrue(AnsibleDistribution.objects.filter(base_path='new-remote').exists())

    def test_associate_existing_entities(self):
        out = StringIO()

        existing_remote = CollectionRemote.objects.create(name='existing-remote')
        existing_repository = AnsibleRepository.objects.create(
            name='existing-repo', remote=existing_remote
        )
        AnsibleDistribution.objects.create(
            name='existing-distro', base_path='existing-distro', repository=existing_repository
        )

        call_command(
            'create-remote',
            'new-remote-with-existing-entities',
            'https://new.remote/v3/collections/',
            '--repository', 'existing-repo',
            '--distribution', 'existing-distro',
            stdout=out
        )

        self.assertIn(
            'Created new CollectionRemote new-remote-with-existing-entities',
            out.getvalue()
        )
        self.assertIn('Associated existing Repository existing-repo', out.getvalue())
        self.assertIn('Associated existing Distribution existing-distro', out.getvalue())

        remote = CollectionRemote.objects.get(name='new-remote-with-existing-entities')
        repo = AnsibleRepository.objects.get(name='existing-repo')
        distro = AnsibleDistribution.objects.get(name='existing-distro')

        self.assertEqual(distro.remote.pk, remote.pk)
        self.assertEqual(distro.remote.pk, repo.remote.pk)
        self.assertEqual(distro.repository.pk, repo.pk)
        self.assertEqual(repo.pk, remote.repository_set.first().pk)

    def test_invalid_url(self):
        with self.assertRaisesMessage(
            CommandError, "url should end with '/'"
        ):
            call_command('create-remote', 'invalid-remote', 'https://invalid.url/foo')

    def test_cannot_create_community_wo_requirements_file(self):
        with self.assertRaisesMessage(
            CommandError,
            'Syncing content from community domains without specifying a '
            'requirements file is not allowed.'
        ):
            call_command('create-remote', 'community', 'https://galaxy.ansible.com/v3/collections/')

    def test_create_community_remote_with_requirements_file_str(self):
        out = StringIO()
        call_command(
            'create-remote',
            'My New Remote',
            'https://galaxy.ansible.com/v3/collections/',
            '-r', REQUIREMENTS_FILE_STR,
            stdout=out
        )

        self.assertIn('Created new CollectionRemote my-new-remote', out.getvalue())
        self.assertIn('Created new Repository my-new-remote', out.getvalue())
        self.assertIn('Created new Distribution my-new-remote', out.getvalue())

        self.assertEqual(
            CollectionRemote.objects.get(name='my-new-remote').requirements_file,
            REQUIREMENTS_FILE_STR
        )

    def test_invalid_requirements_file(self):
        with self.assertRaisesMessage(
            CommandError,
            "Error parsing requirements_file"
        ):
            call_command('create-remote', 'foo', 'https://bar.com/', '-r', 'invalid_file_str')

    def test_load_requirements_file_from_path(self):
        out = StringIO()

        with tempfile.NamedTemporaryFile(mode='w+', suffix='.yaml') as requirements_file:
            requirements_file.write(REQUIREMENTS_FILE_STR)
            requirements_file.seek(0)

            call_command(
                'create-remote',
                'community-from-file',
                'https://galaxy.ansible.com/v3/collections/',
                '-r', requirements_file.name,
                stdout=out
            )

            self.assertIn('Created new CollectionRemote community', out.getvalue())
            self.assertIn('Created new Repository community', out.getvalue())
            self.assertIn('Created new Distribution community', out.getvalue())

            self.assertIn('requirements_file loaded and parsed', out.getvalue())



from django.shortcuts import get_object_or_404
from django_filters import filters
from django_filters.rest_framework import filterset, DjangoFilterBackend

from rest_framework import mixins

from galaxy_ng.app.models import auth as auth_models
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.api.ui import serializers, versioning
from galaxy_ng.app.api import base as api_base


class UserFilter(filterset.FilterSet):
    sort = filters.OrderingFilter(
        fields=(
            ('username', 'username'),
            ('email', 'email'),
            ('first_name', 'first_name'),
            ('last_name', 'last_name'),
            ('date_joined', 'date_joined')
        )
    )

    class Meta:
        model = auth_models.User
        fields = {
            'username': ['exact', 'contains', 'startswith'],
            'email': ['exact', 'contains', 'startswith'],
            'first_name': ['exact', 'contains', 'startswith'],
            'last_name': ['exact', 'contains', 'startswith'],
            'date_joined': ['exact'],
            'groups__name': ['exact'],
            'groups': ['exact']
        }


class UserViewSet(
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    api_base.GenericViewSet,
):
    serializer_class = serializers.UserSerializer
    model = auth_models.User
    queryset = auth_models.User.objects.all()
    filter_backends = (DjangoFilterBackend,)
    filterset_class = UserFilter
    permission_classes = [access_policy.UserAccessPolicy]
    versioning_class = versioning.UIVersioning


class CurrentUserViewSet(
    api_base.GenericViewSet,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
):
    serializer_class = serializers.CurrentUserSerializer
    model = auth_models.User
    permission_classes = [access_policy.MyUserAccessPolicy]
    versioning_class = versioning.UIVersioning

    def get_object(self):
        return get_object_or_404(self.model, pk=self.request.user.pk)



import urllib

from django.test import override_settings
from pulp_ansible.app.models import (
    AnsibleDistribution,
    AnsibleRepository,
    Collection,
    CollectionVersion,
    CollectionRemote
)
from galaxy_ng.app import models
from galaxy_ng.app.constants import DeploymentMode
from .base import BaseTestCase, get_current_ui_url


def _create_repo(name, **kwargs):
    repo = AnsibleRepository.objects.create(name=name, **kwargs)
    AnsibleDistribution.objects.create(name=name, base_path=name, repository=repo)
    return repo


def _create_remote(name, url, **kwargs):
    return CollectionRemote.objects.create(
        name=name, url=url, **kwargs
    )


def _get_create_version_in_repo(namespace, collection, version, repo):
    collection_version, _ = CollectionVersion.objects.get_or_create(
        namespace=namespace,
        name=collection.name,
        collection=collection,
        version=version,
    )
    qs = CollectionVersion.objects.filter(pk=collection_version.pk)
    with repo.new_version() as new_version:
        new_version.add_content(qs)


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class TestUiCollectionVersionViewSet(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.versions_url = get_current_ui_url('collection-versions-list')
        self.namespace = models.Namespace.objects.create(name='my_namespace')
        self.collection = Collection.objects.create(namespace=self.namespace, name='my_collection')

        _get_create_version_in_repo(
            self.namespace, self.collection, '1.1.1', _create_repo(name='repo1'))
        _get_create_version_in_repo(
            self.namespace, self.collection, '1.1.2', _create_repo(name='repo2'))

    def _versions_url_with_params(self, query_params):
        return self.versions_url + '?' + urllib.parse.urlencode(query_params)

    def test_no_filters(self):
        response = self.client.get(self.versions_url)
        self.assertEqual(response.data['meta']['count'], 2)

    def test_repo_filter(self):
        url = self._versions_url_with_params({'repository': 'repo_dne'})
        response = self.client.get(url)
        self.assertEqual(response.data['meta']['count'], 0)

        url = self._versions_url_with_params({'repository': 'repo1'})
        response = self.client.get(url)
        self.assertEqual(response.data['meta']['count'], 1)
        self.assertEqual(response.data['data'][0]['version'], '1.1.1')

    def test_multiple_filters(self):
        url = self._versions_url_with_params({
            'namespace': 'namespace_dne',
            'version': '1.1.2',
            'repository': 'repo2',
        })
        response = self.client.get(url)
        self.assertEqual(response.data['meta']['count'], 0)

        url = self._versions_url_with_params({
            'namespace': 'my_namespace',
            'version': '1.1.2',
            'repository': 'repo2',
        })
        response = self.client.get(url)
        self.assertEqual(response.data['meta']['count'], 1)
        self.assertEqual(response.data['data'][0]['version'], '1.1.2')

    def test_sort_and_repo_list(self):
        url = self._versions_url_with_params({'sort': 'pulp_created'})
        response = self.client.get(url)
        self.assertEqual(response.data['data'][0]['version'], '1.1.1')
        self.assertEqual(response.data['data'][0]['repository_list'], ['repo1'])

        url = self._versions_url_with_params({'sort': '-pulp_created'})
        response = self.client.get(url)
        self.assertEqual(response.data['data'][0]['version'], '1.1.2')
        self.assertEqual(response.data['data'][0]['repository_list'], ['repo2'])


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class TestUiCollectionViewSet(BaseTestCase):
    def setUp(self):
        super().setUp()
        namespace_name = 'my_namespace'
        collection1_name = 'collection1'
        collection2_name = 'collection2'

        self.repo1 = _create_repo(name='repo1')
        self.repo2 = _create_repo(name='repo2')
        self.repo3 = _create_repo(name='repo3')
        self.namespace = models.Namespace.objects.create(name=namespace_name)
        self.collection1 = Collection.objects.create(
            namespace=self.namespace, name=collection1_name)
        self.collection2 = Collection.objects.create(
            namespace=self.namespace, name=collection2_name)

        _get_create_version_in_repo(self.namespace, self.collection1, '1.0.0', self.repo1)
        _get_create_version_in_repo(self.namespace, self.collection1, '1.0.1', self.repo1)
        _get_create_version_in_repo(self.namespace, self.collection2, '2.0.0', self.repo1)
        _get_create_version_in_repo(self.namespace, self.collection1, '1.0.0', self.repo2)

        self.repo1_list_url = get_current_ui_url(
            'collections-list', kwargs={'path': 'repo1'})
        self.repo2_list_url = get_current_ui_url(
            'collections-list', kwargs={'path': 'repo2'})
        self.repo3_list_url = get_current_ui_url(
            'collections-list', kwargs={'path': 'repo3'})
        self.repo1_collection1_detail_url = get_current_ui_url(
            'collections-detail',
            kwargs={
                'path': 'repo1',
                'namespace': namespace_name,
                'name': collection1_name})

    def test_list_count(self):
        response = self.client.get(self.repo1_list_url)
        self.assertEqual(response.data['meta']['count'], 2)

        response = self.client.get(self.repo2_list_url)
        self.assertEqual(response.data['meta']['count'], 1)

        response = self.client.get(self.repo3_list_url)
        self.assertEqual(response.data['meta']['count'], 0)

    def test_list_latest_version(self):
        response = self.client.get(self.repo1_list_url)
        c1 = next(i for i in response.data['data'] if i['name'] == self.collection1.name)
        self.assertEqual(c1['latest_version']['version'], '1.0.1')

        c2 = next(i for i in response.data['data'] if i['name'] == self.collection2.name)
        self.assertEqual(c2['latest_version']['version'], '2.0.0')

        response = self.client.get(self.repo2_list_url)
        c1 = next(i for i in response.data['data'] if i['name'] == self.collection1.name)
        self.assertEqual(c1['latest_version']['version'], '1.0.0')

    def test_detail_latest_version(self):
        response = self.client.get(self.repo1_collection1_detail_url)
        self.assertEqual(response.data['latest_version']['version'], '1.0.1')


@override_settings(GALAXY_DEPLOYMENT_MODE=DeploymentMode.STANDALONE.value)
class TestUiCollectionRemoteViewSet(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.remote_data = {
            "name": "rh-certified",
            "url": "https://cloud.redhat.com/api/automation-hub/",
        }
        self.remote = CollectionRemote.objects.get(name=self.remote_data["name"])
        self.repository = AnsibleRepository.objects.get(name=self.remote_data["name"])

    def test_get_remotes(self):
        # data migration created 2 remotes
        response = self.client.get(get_current_ui_url('remotes-list'))
        self.assertEqual(response.data['meta']['count'], 2)

        for key, value in self.remote_data.items():
            self.assertEqual(response.data['data'][1][key], value)

        repository = response.data['data'][1]['repositories'][0]

        self.assertEqual(repository['name'], self.remote_data["name"])
        self.assertEqual(repository['distributions'][0]['base_path'], self.remote_data["name"])

        # token is not visible in a GET
        self.assertNotIn('token', response.data['data'][1])



from django.db import migrations

MIGRATE_COLLECTIONS_QUERY = """
CREATE EXTENSION IF NOT EXISTS "uuid-ossp";

CREATE FUNCTION _new_repository_version(repository_id UUID) RETURNS UUID AS $$
DECLARE
    result UUID;
BEGIN
     INSERT INTO core_repositoryversion (
        pulp_id, pulp_created, pulp_last_updated, number,
        complete, base_version_id, repository_id
    ) VALUES (
        uuid_generate_v4(), now(), now(),
        (SELECT next_version FROM core_repository WHERE pulp_id = repository_id),
        TRUE, NULL, repository_id
    ) RETURNING pulp_id INTO result;

    UPDATE core_repository SET next_version = next_version + 1 WHERE pulp_id = repository_id;

    RETURN result;
END;
$$ LANGUAGE plpgsql;

CREATE FUNCTION _move_collection_versions_by_certification(
    from_repo_name TEXT, to_repo_name TEXT,  filter TEXT
) RETURNS void AS
$$
DECLARE
    from_repo UUID := (
        SELECT pulp_id FROM core_repository WHERE name = from_repo_name
    );
    to_repo UUID := (
        SELECT pulp_id FROM core_repository WHERE name = to_repo_name
    );
    from_version UUID;
    to_version   UUID;
BEGIN
    IF NOT EXISTS(SELECT 1 FROM ansible_collectionversion cv WHERE cv.certification = filter) THEN
        RAISE NOTICE 'Nothing to migrate from %, to %, filter %', 
            from_repo_name, to_repo_name, filter;
        RETURN;
    END IF;

    from_version := (SELECT _new_repository_version(from_repo));
    to_version := (SELECT _new_repository_version(to_repo));

    UPDATE core_repositorycontent SET version_removed_id = from_version
    WHERE version_removed_id IS NULL AND content_id IN (
        SELECT cv.content_ptr_id
        FROM ansible_collectionversion cv
        WHERE cv.certification = filter
    );

    INSERT INTO core_repositorycontent (
        pulp_id, pulp_created, pulp_last_updated, content_id,
        repository_id, version_added_id, version_removed_id
    ) SELECT
        uuid_generate_v4(), now(), now(), cv.content_ptr_id,
        to_repo, to_version, NULL
    FROM ansible_collectionversion cv
    WHERE cv.certification = filter;

END ;
$$ LANGUAGE plpgsql;

LOCK TABLE
    core_repository,
    core_repositoryversion,
    core_repositorycontent,
    ansible_collectionversion
IN ACCESS EXCLUSIVE MODE;
    
SELECT _move_collection_versions_by_certification(
    'published', 'rejected', 'not_certified'
);
               
SELECT _move_collection_versions_by_certification(
    'published', 'staging', 'needs_review'
);

DROP FUNCTION _move_collection_versions_by_certification(TEXT, TEXT, TEXT);
DROP FUNCTION _new_repository_version(UUID);
"""


class Migration(migrations.Migration):

    dependencies = [
        ('galaxy', '0011_collection_import_task_id_fk'),
    ]

    operations = [
        migrations.RunSQL(sql=MIGRATE_COLLECTIONS_QUERY, elidable=True),
    ]



import logging

from django.conf import settings
from django.apps import apps

from rest_framework import status as http_code

from galaxy_ng.app.models import auth as auth_models
from . import base

from .synclist_base import BaseSyncListViewSet, ACCOUNT_SCOPE

log = logging.getLogger(__name__)
logging.getLogger().setLevel(logging.DEBUG)

log.info("settings.FIXTURE_DIRS(module scope): %s", settings.FIXTURE_DIRS)


class TestUiMySyncListViewSet(BaseSyncListViewSet):
    def setUp(self):
        super().setUp()

        log.info("self.fixtures2: %s", self.fixtures)
        log.info("settings.FIXTURE_DIRS2: %s", settings.FIXTURE_DIRS)

        self.user = auth_models.User.objects.create_user(username="test1", password="test1-secret")
        self.group = self._create_group_with_synclist_perms(
            ACCOUNT_SCOPE, "test1_group", users=[self.user]
        )
        self.user.save()
        self.group.save()

        self.user.groups.add(self.group)
        self.user.save()

        self.synclist_name = "test_synclist"
        self.synclist = self._create_synclist(
            name=self.synclist_name,
            repository=self.repo,
            upstream_repository=self.default_repo,
            groups=[self.group],
        )

        self.client.force_authenticate(user=self.user)

    def test_my_synclist_create(self):
        galaxy_app_config = apps.get_app_config("galaxy")
        log.info("gac module: %s", galaxy_app_config.module)
        log.info("gac path: %s", galaxy_app_config.path)

        post_data = {
            "repository": self.repo.pulp_id,
            "collections": [],
            "namespaces": [],
            "policy": "include",
            "groups": [
                {
                    "id": self.group.id,
                    "name": self.group.name,
                    "object_permissions": self.default_owner_permissions,
                },
            ],
        }

        synclists_url = base.get_current_ui_url("my-synclists-list")

        response = self.client.post(synclists_url, post_data, format="json")

        log.debug("response: %s", response)
        log.debug("response.data: %s", response.data)

        # synclist create is not allowed via my-synclist viewset
        self.assertEqual(response.status_code, http_code.HTTP_403_FORBIDDEN, msg=response.data)

    def test_my_synclist_update(self):
        ns_name = "unittestnamespace1"
        ns = self._create_namespace(ns_name, groups=[self.group])
        ns.save()

        post_data = {
            "repository": self.repo.pulp_id,
            "collections": [],
            "namespaces": [ns_name],
            "policy": "include",
            "groups": [
                {
                    "id": self.group.id,
                    "name": self.group.name,
                    "object_permissions": self.default_owner_permissions,
                },
            ],
        }

        synclists_detail_url = base.get_current_ui_url(
            "my-synclists-detail", kwargs={"pk": self.synclist.id}
        )

        response = self.client.patch(synclists_detail_url, post_data, format="json")

        log.debug("response: %s", response)
        log.debug("response.data: %s", response.data)

        self.assertEqual(response.status_code, http_code.HTTP_200_OK)
        self.assertIn("name", response.data)
        self.assertIn("repository", response.data)
        self.assertEqual(response.data["name"], self.synclist_name)
        self.assertEqual(response.data["policy"], "include")

        # Sort permission list for comparison
        response.data["groups"][0]["object_permissions"].sort()
        self.default_owner_permissions.sort()
        self.assertEqual(
            response.data["groups"],
            [
                {
                    "name": self.group.name,
                    "id": self.group.id,
                    "object_permissions": self.default_owner_permissions,
                }
            ],
        )

    def test_my_synclist_list(self):
        synclists_url = base.get_current_ui_url("my-synclists-list")
        log.debug("synclists_url: %s", synclists_url)

        response = self.client.get(synclists_url)

        log.debug("response: %s", response)
        data = response.data["data"]

        self.assertIsInstance(data, list)
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]["name"], self.synclist_name)
        self.assertEqual(data[0]["policy"], "exclude")
        self.assertEqual(data[0]["repository"], self.repo.pulp_id)

    def test_my_synclist_detail(self):
        synclists_detail_url = base.get_current_ui_url(
            "my-synclists-detail", kwargs={"pk": self.synclist.id}
        )

        log.debug("synclists_detail_url: %s", synclists_detail_url)

        response = self.client.get(synclists_detail_url)

        log.debug("response: %s", response)

        self.assertEqual(response.status_code, http_code.HTTP_200_OK)
        self.assertIn("name", response.data)
        self.assertIn("repository", response.data)
        self.assertEqual(response.data["name"], self.synclist_name)
        self.assertEqual(response.data["policy"], "exclude")
        self.assertEqual(response.data["collections"], [])
        self.assertEqual(response.data["namespaces"], [])

    def test_my_synclist_delete(self):
        synclists_detail_url = base.get_current_ui_url(
            "my-synclists-detail", kwargs={"pk": self.synclist.id}
        )

        log.debug("delete url: %s", synclists_detail_url)

        response = self.client.delete(synclists_detail_url)

        log.debug("delete response: %s", response)

        self.assertEqual(response.status_code, http_code.HTTP_403_FORBIDDEN)



import logging

from pulpcore.plugin import viewsets as pulp_core_viewsets

from galaxy_ng.app.api.base import LocalSettingsMixin
from galaxy_ng.app.api.v3.serializers import TaskSerializer, TaskSummarySerializer
from galaxy_ng.app.access_control import access_policy

log = logging.getLogger(__name__)


class TaskViewSet(LocalSettingsMixin, pulp_core_viewsets.TaskViewSet):
    permission_classes = [access_policy.TaskAccessPolicy]
    serializer_class = TaskSerializer

    def get_serializer_class(self):
        if self.action == 'list':
            return TaskSummarySerializer
        else:
            return self.serializer_class



from django.conf import settings
from django.urls import include, path

from . import views
from .ui import urls as ui_urls
from .v3 import urls as v3_urls

DEFAULT_DISTRIBUTION_BASE_PATH = settings.GALAXY_API_DEFAULT_DISTRIBUTION_BASE_PATH.strip('/')

app_name = "api"


v3_urlpatterns = [
    path("", include(v3_urls.auth_urls)),
    path("", include(v3_urls.namespace_urls)),

    # Set an instance of the v3 urls/viewsets at the same
    # url path as the former galaxy_api.api.v3.viewsets
    # ie (/api/automation-hub/v3/collections etc
    # and pass in path param 'path' that actually means
    # the base_path of the Distribution. For the default case,
    # use the hard coded 'default' distro (ie, 'automation-hub')
    path(
        "",
        include((v3_urls, app_name), namespace='default-content'),
        {
            'no_path_specified': True,
            'path': DEFAULT_DISTRIBUTION_BASE_PATH,
        }
    ),
]

content_v3_urlpatterns = [
    # A set of galaxy v3 API urls per Distribution.base_path
    # which is essentially per Repository
    path("<str:path>/v3/",
         # include((v3_urls, app_name))),
         include(v3_urls)),
    path("<str:path>/v3/", include(v3_urls.sync_urls)),
]

content_urlpatterns = [
    path("content/",
         include((content_v3_urlpatterns, app_name), namespace="v3")),

    path("content/<str:path>/",
         views.ApiRootView.as_view(),
         name="root"),

    # This can be removed when ansible-galaxy stops appending '/api' to the
    # urls.
    path("content/<str:path>/api/",
         views.ApiRedirectView.as_view(),
         name="api-redirect",
         kwargs={"reverse_url_name": "galaxy:api:content:root"}),
]

urlpatterns = [
    path("v3/", include((v3_urlpatterns, app_name), namespace="v3")),

    path("_ui/", include((ui_urls, app_name), namespace="ui")),

    path("", include((content_urlpatterns, app_name), namespace='content')),

    path("",
         views.ApiRootView.as_view(),
         name="root",
         ),

    # This path is to redirect requests to /api/automation-hub/api/
    # to /api/automation-hub/

    # This is a workaround for https://github.com/ansible/ansible/issues/62073.
    # ansible-galaxy in ansible 2.9 always appends '/api' to any configured
    # galaxy server urls to try to find the API root. So add a redirect from
    # "/api" to actual API root at "/".

    # This can be removed when ansible-galaxy stops appending '/api' to the
    # urls.
    path("api/",
         views.ApiRedirectView.as_view(),
         name="api-redirect",
         kwargs={"reverse_url_name": "galaxy:api:root"}),
]



from rest_framework import serializers


class GroupSummarySerializer(serializers.Serializer):
    id = serializers.IntegerField(required=False)
    name = serializers.CharField(required=False, allow_blank=False)



from rest_framework.serializers import CharField

from .base import Serializer


class LoginSerializer(Serializer):
    username = CharField(required=True)
    password = CharField(required=True, style={'input_type': 'password'})



from django.db.models import Exists, OuterRef, Q
from django.core.exceptions import ObjectDoesNotExist
from django.http import Http404
from django.shortcuts import get_object_or_404
from django_filters import filters
from django_filters.rest_framework import filterset, DjangoFilterBackend, OrderingFilter
from drf_spectacular.utils import extend_schema
from pulp_ansible.app.galaxy.v3 import views as pulp_ansible_galaxy_views
from pulp_ansible.app import viewsets as pulp_ansible_viewsets
from pulp_ansible.app.models import (
    AnsibleCollectionDeprecated,
    AnsibleDistribution,
    CollectionVersion,
    Collection,
    CollectionRemote,
)
from pulp_ansible.app.models import CollectionImport as PulpCollectionImport
from rest_framework import mixins
from rest_framework.exceptions import NotFound
from rest_framework.response import Response
import semantic_version

from galaxy_ng.app.api import base as api_base
from galaxy_ng.app.access_control import access_policy
from galaxy_ng.app.api.ui import serializers, versioning
from galaxy_ng.app.api.v3.serializers.sync import CollectionRemoteSerializer


class CollectionByCollectionVersionFilter(pulp_ansible_viewsets.CollectionVersionFilter):
    """pulp_ansible CollectionVersion filter for Collection viewset."""
    versioning_class = versioning.UIVersioning
    keywords = filters.CharFilter(field_name="keywords", method="filter_by_q")
    deprecated = filters.BooleanFilter()


class CollectionViewSet(
    api_base.GenericViewSet,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin,
    pulp_ansible_galaxy_views.AnsibleDistributionMixin,
):
    """Viewset that uses CollectionVersion's within distribution to display data for Collection's.

    Collection list is filterable by FilterSet and includes latest CollectionVersion.

    Collection detail includes CollectionVersion that is latest or via query param 'version'.
    """
    versioning_class = versioning.UIVersioning
    filterset_class = CollectionByCollectionVersionFilter
    permission_classes = [access_policy.CollectionAccessPolicy]

    def get_queryset(self):
        """Returns a CollectionVersions queryset for specified distribution."""
        path = self.kwargs.get('path')
        if path is None:
            raise Http404("Distribution base path is required")

        versions = CollectionVersion.objects.filter(pk__in=self._distro_content).values_list(
            "collection_id",
            "version",
        )

        collection_versions = {}
        for collection_id, version in versions:
            value = collection_versions.get(str(collection_id))
            if not value or semantic_version.Version(version) > semantic_version.Version(value):
                collection_versions[str(collection_id)] = version

        if not collection_versions.items():
            return CollectionVersion.objects.none().annotate(
                # AAH-122: annotated fields must exist in all the returned querysets
                #          in order for filters to work.
                deprecated=Exists(AnsibleCollectionDeprecated.objects)
            )

        query_params = Q()
        for collection_id, version in collection_versions.items():
            query_params |= Q(collection_id=collection_id, version=version)

        deprecated_query = AnsibleCollectionDeprecated.objects.filter(
            collection=OuterRef("collection"), repository_version=self._repository_version
        )
        version_qs = CollectionVersion.objects.select_related("collection").filter(query_params)
        version_qs = version_qs.annotate(deprecated=Exists(deprecated_query))
        return version_qs

    def get_object(self):
        """Return CollectionVersion object, latest or via query param 'version'."""
        version = self.request.query_params.get('version', None)

        if not version:
            queryset = self.get_queryset()
            return get_object_or_404(
                queryset, namespace=self.kwargs["namespace"], name=self.kwargs["name"]
            )

        return get_object_or_404(
            CollectionVersion.objects.all(),
            pk__in=self._distro_content,
            namespace=self.kwargs["namespace"],
            name=self.kwargs["name"],
            version=version,
        )

    def get_serializer_class(self):
        if self.action == 'list':
            return serializers.CollectionListSerializer
        else:
            return serializers.CollectionDetailSerializer


class CollectionVersionFilter(filterset.FilterSet):
    repository = filters.CharFilter(field_name='repository', method='repo_filter')
    versioning_class = versioning.UIVersioning

    def repo_filter(self, queryset, name, value):
        try:
            distro = AnsibleDistribution.objects.get(base_path=value)
            repository_version = distro.repository.latest_version()
            return queryset.filter(pk__in=repository_version.content)
        except ObjectDoesNotExist:
            return CollectionVersion.objects.none()

    sort = OrderingFilter(
        fields=(
            ('pulp_created', 'pulp_created'),
            ('namespace', 'namespace'),
            ('name', 'collection'),
            ('version', 'version'),
        )
    )

    class Meta:
        model = CollectionVersion
        fields = ['namespace', 'name', 'version', 'repository']


class CollectionVersionViewSet(api_base.GenericViewSet):
    lookup_url_kwarg = 'version'
    lookup_value_regex = r'[0-9a-z_]+/[0-9a-z_]+/[0-9A-Za-z.+-]+'
    queryset = CollectionVersion.objects.all()
    serializer_class = serializers.CollectionVersionSerializer
    filterset_class = CollectionVersionFilter
    versioning_class = versioning.UIVersioning

    permission_classes = [access_policy.CollectionAccessPolicy]

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        page = self.paginate_queryset(queryset)
        serializer = self.get_serializer(page, many=True)
        return self.get_paginated_response(serializer.data)

    @extend_schema(summary="Retrieve collection version",
                   responses={200: serializers.CollectionVersionDetailSerializer})
    def retrieve(self, request, *args, **kwargs):
        namespace, name, version = self.kwargs['version'].split('/')
        try:
            collection_version = CollectionVersion.objects.get(
                namespace=namespace,
                collection=Collection.objects.get(namespace=namespace, name=name),
                version=version,
            )
        except ObjectDoesNotExist:
            raise NotFound(f'Collection version not found for: {self.kwargs["version"]}')
        serializer = serializers.CollectionVersionDetailSerializer(collection_version)
        return Response(serializer.data)


class CollectionImportFilter(filterset.FilterSet):
    namespace = filters.CharFilter(field_name='galaxy_import__namespace__name')
    name = filters.CharFilter(field_name='galaxy_import__name')
    keywords = filters.CharFilter(field_name='galaxy_import__name', lookup_expr='icontains')
    state = filters.CharFilter(field_name='task__state')
    version = filters.CharFilter(field_name='galaxy_import__version')
    created = filters.DateFilter(field_name='galaxy_import__created_at')
    versioning_class = versioning.UIVersioning

    sort = OrderingFilter(
        fields=(('galaxy_import__created_at', 'created'),)
    )

    class Meta:
        model = PulpCollectionImport
        fields = ['namespace',
                  'name',
                  'keywords',
                  'state',
                  'version']


class CollectionImportViewSet(api_base.GenericViewSet,
                              pulp_ansible_galaxy_views.CollectionImportViewSet):
    lookup_field = 'task_id'
    queryset = PulpCollectionImport.objects.prefetch_related("task", "galaxy_import").all()
    serializer_class = serializers.ImportTaskListSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_class = CollectionImportFilter

    versioning_class = versioning.UIVersioning

    ordering_fields = ('created',)

    permission_classes = [access_policy.CollectionAccessPolicy]

    def get_serializer_class(self):
        if self.action == 'list':
            return serializers.ImportTaskListSerializer
        else:
            return serializers.ImportTaskDetailSerializer

    def list(self, request, *args, **kwargs):
        qs = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(qs)
        serializer = self.get_serializer(page, many=True)

        return self.get_paginated_response(serializer.data)

    @extend_schema(summary="Retrieve collection import",
                   responses={200: serializers.ImportTaskDetailSerializer})
    def retrieve(self, request, *args, **kwargs):
        task = self.get_object()
        data = serializers.ImportTaskDetailSerializer(task).data
        return Response(data)


class CollectionRemoteViewSet(api_base.ModelViewSet):
    queryset = CollectionRemote.objects.all().order_by('name')
    serializer_class = CollectionRemoteSerializer

    permission_classes = [access_policy.CollectionRemoteAccessPolicy]
